/* pftest.c
 * ============================================================
 * Demo-Programm f|r formatierte Ausgabe (95/04/28)
 * Copyright 1995: Martin Weitzel, Darmstadt
 * ============================================================
*/

#include <setjmp.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS (0)	/* fehlt manchmal in <stdlib.h> */
#endif
#ifndef EXIT_FAILURE
#define EXIT_FAILURE (1)	/* fehlt manchmal in <stdlib.h> */
#endif

#define NONE	(-1)	/* keine Spezifikation */
#define POINTER	100	/* allgemeiner Zeiger */
#define CHAR	200	/* einzelnes Zeichen */
#define STRING	300	/* C-Zeichenkette */
#define INT	400	/* Ganzzahl */
#define LONG	500	/* lange Ganzzahl */
#define DOUBLE	600	/* lange Gleitpunktzahl */
#define LDOUBLE	700	/* extra lange Gleitpunktzahl */

#define argtype(t) ((t)/100)	/* Argumenttyp */
#define xpcount(t) ((t)%100)	/* Extras (wg. '*') */

jmp_buf badinput;	/* fuer Ruecksprungpunkt */

const char *sget(int sc)	/* -- Einlesen eines String */
{
	static char old[100];
	if (sc && scanf(" %99[^\n]", old) != 1)
		longjmp(badinput, STRING);
	return old;
}
/*
 * Hilfsmakro zur Definition von sechs weiteren,
 * sehr aehnlichen Einlesefunktionen ...
*/
#define	DEFGET(name, type, scanfmt, errcode)\
	type name(int sc)\
	{\
		static type old;\
		if (sc == 1 && scanf(scanfmt, &old) != 1)\
			longjmp(badinput, errcode);\
		return old;\
	}
/*
 * ... und Definition dieser Funktionen:
*/
	DEFGET(	g1,	int,		"%d",	1	)
	DEFGET(	g2,	int,		"%d",	2	)
	DEFGET(	cget,	char,		"%c",	CHAR	)
	DEFGET(	iget,	int,		"%i",	INT	)
	DEFGET(	lget,	long,		"%li",	LONG	)
	DEFGET(	dget,	double,		"%lf",	DOUBLE	)
#ifndef	NO_LONGDOUBLE
	DEFGET(	ldget,	long double,	"%Lf",	LDOUBLE	)
#endif
#undef	DEFGET

void process(		/* -- ggf. Werte lesen, dann ausgeben */
	const char *f,		/* Formatstring */
	int t,			/* Umwandlungsart */
	int n			/* wirklich lesen? */
) {
	switch (t) {
	case 0+CHAR:
		(void) printf(f, cget(n));
		break;
	case 1+CHAR:
		(void) printf(f, g1(n), cget(n));
		break;
	case 2+CHAR:
		(void) printf(f, g1(n), g2(n), cget(n));
		break;
	case 0+STRING:
		(void) printf(f, sget(n));
		break;
	case 1+STRING:
		(void) printf(f, g1(n), sget(n));
		break;
	case 2+STRING:
		(void) printf(f, g1(n), g2(n), sget(n));
		break;
	case 0+INT:
		(void) printf(f, iget(n));
		break;
	case 1+INT:
		(void) printf(f, g1(n), iget(n));
		break;
	case 2+INT:
		(void) printf(f, g1(n), g2(n), iget(n));
		break;
	case 0+LONG:
		(void) printf(f, lget(n));
		break;
	case 1+LONG:
		(void) printf(f, g1(n), lget(n));
		break;
	case 2+LONG:
		(void) printf(f, g1(n), g2(n), lget(n));
		break;
	case 0+DOUBLE:
		(void) printf(f, dget(n));
		break;
	case 1+DOUBLE:
		(void) printf(f, g1(n), dget(n));
		break;
	case 2+DOUBLE:
		(void) printf(f, g1(n), g2(n), dget(n));
		break;
#ifndef	NO_LONGDOUBLE
	case 0+LDOUBLE:
		(void) printf(f, dget(n));
		break;
	case 1+LDOUBLE:
		(void) printf(f, g1(n), ldget(n));
		break;
	case 2+LDOUBLE:
		(void) printf(f, g1(n), g2(n), ldget(n));
		break;
#endif
	}
	while (n && !feof(stdin) && getc(stdin) != '\n')
		; /*empty*/
}

int getfspec(		/* -- Umwandlungsspezifikation */
	const char *f		/* Formatstring */
) {
	if ((f = strchr(f, '%')) != (char *)0) {
		int x = 0;
		while (*++f) {
			switch (*f) {
			case '*':
				++x;
				continue;
			case 'c':
				return x+CHAR;
			case 's':
				return x+STRING;
			case 'd': case 'i': case 'u':
			case 'o': case 'x': case 'X':
				return f[-1] == 'l'
					? x+LONG : x+INT;
			case 'e': case 'E': case 'f':
			case 'g': case 'G':
				return f[-1] == 'L'
					? x+LDOUBLE : x+DOUBLE;
			case 'p':
				return x+POINTER;
			}
		}
	}
	return NONE;
}

const char *PROG;	/* Programmname */

/*PRINTFLIKE1*/
void failure(		/* -- Abbruch bei fatalen Fehlern */
	const char *message,	/* Meldungstext */
	...			/* Meldungsparameter */
) {
	va_list ap;
	(void) fprintf(stderr, "%s: ", PROG);
	va_start(ap, message);
	(void) vfprintf(stderr, message, ap);
	va_end(ap);
	(void) fprintf(stderr, "\n");
	exit(EXIT_FAILURE);
	/*NOTREACHED*/
}

void converr(		/* -- Umwandlungsfehler melden */
	const char *w		/* Datentyp */
) {
	(void) printf("conversion error (%s)\n", w);
}

int main(		/* -- Hauptprogramm */
	int argc,		/* Argumentzahl */
	char *argv[]		/* Argumentvektor */
) {
	int i, ec, fspec, lastfspec;

	PROG = argv[0];

	if (argc < 2) {
		failure("argument(s) required");
		/*NOTREACHED*/
	}
	for (i = 1; i < argc; ++i) {
		fspec = getfspec(argv[i]);
		if (i == 1)
			lastfspec = fspec;
		else if (lastfspec != fspec) {
			failure("not compatible (%s : %s)\n",
				argv[i-1], argv[i]);
		}
		switch (argtype(fspec)) {
		case POINTER:
#ifdef	NO_LONGDOUBLE
		case LDOUBLE:
#endif
			failure("unsupported conversion");
			/*NOTREACHED*/
		case NONE:
			failure("missing conversion");
			/*NOTREACHED*/
		default:
			if (xpcount(fspec) > 2)
				failure("format syntax error");
		}
	}

	ec = setjmp(badinput);
	if (ec && !feof(stdin)) {
		switch (ec) {
		case 1:
			converr("1-st int");
			break;
		case 2:
			converr("2-nd int");
			break;
		case CHAR:
			converr("char");
			break;
		case STRING:
			converr("string");
			break;
		case INT:
			converr("int");
			break;
		case LONG:
			converr("long");
			break;
		case DOUBLE:
			converr("double");
			break;
		case LDOUBLE:
			converr("long double");
			break;
		}
		(void) scanf("%*[^\n]");
	}
	while (ungetc(getc(stdin), stdin) != EOF) {
		for (i = 1; i < argc; ++i) {
			process(argv[i], fspec, i == 1);
			(void) printf("|");
		}
		(void) printf("\n");
	}
	return EXIT_SUCCESS;
}
