/* prfile.c
 ==============================================================
 Implementation: Modul PRFILE (95/04/27)
 Copyright 1995: Martin Weitzel, Darmstadt
 ==============================================================
*/

#include "prfile.h"

#include <assert.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

PRFILE *prfopen(	/* -- PRFILE oeffnen */
	const char *name,	/* Dateiname */
	const char *mode,	/* Oeffnungsmodus */
	void (*panic)(void)	/* Fehlerbehandlung */
) {
	PRFILE *prfp;
	assert(name != (const char *)0);
	assert(mode != (const char *)0);
	assert(panic != (void (*)(void))0);
	prfp = (PRFILE *)malloc(sizeof(PRFILE));
	if (prfp != (PRFILE *)0) {
		prfp->file = fopen(name, mode);
		if (prfp->file != (FILE *)0)
			prfp->panic = panic;
		else {
			free(prfp);
			prfp = (PRFILE *)0;
		}
	}
	return prfp;
}

void prfprintf(		/* -- formatierte Ausgabe */
	PRFILE *prfp,		/* PRFILE-Pointer */
	const char *fmt,	/* Formatstring */
	...			/* weitere Parameter */
) {
	va_list ap;
	int n;
	assert(prfp != (PRFILE *)0);
	assert(fmt != (const char *)0);
	va_start(ap, fmt);
	n = vfprintf(prfp->file, fmt, ap);
	va_end(ap);
	if (n < 0)
		(*prfp->panic)();
}

void prfflush(		/* -- Puffer leeren */
	PRFILE *prfp		/* PRFILE-Pointer */
) {
	assert(prfp != (PRFILE *)0);
	if (fflush(prfp->file) != 0)
		(*prfp->panic)();
}

void prfclose(		/* -- PRFILE schliessen */
	PRFILE *prfp		/* PRFILE-Pointer */
) {
	assert(prfp != (PRFILE *)0);
	if (fclose(prfp->file) != 0)
		(*prfp->panic)();
	free(prfp);
}

#ifdef TEST
/*
 * Anlegen einer Datei mit Namen gemaess erstem Argument,
 * Schreiben von nummerierten Zeilen gemaes den weiteren
 * Argumenten.  Test des Fehlerabbruchs wie folgt:
 *   a)	Aufruf:	a.out /dev/tty hello world
 *   ->	Ausgabe auf Terminal
 *		1) hello
 *		2) world
 *   b)	Aufruf:	a.out x hello world
 *   ->	wie zuvor, aber Ausgabe in Datei x
 *   c)	Aufruf:	ulimit 0  (++)
 *		a.out x hello world
 *   ->	Fehlerabbruch bei Ausgabe, da max. Dateigroesse 0
 *   d) Aufruf:	a.out
 *   ->	assert-Fehler in prfopen
 * ++:	Da die Wirkung dieses Kommandos nicht rueckgaengig
 *	gemacht werden kann, sollte dieser Test in einer
 *	Sub-Shell ausgefuehrt werden.
*/
int main(		/* -- Hauptprogramm */
	int argc,		/* Argumentzahl */
	char *argv[]		/* Argumentvektor */
) {
	int i;
	PRFILE *prfp = prfopen(argv[1], "w", abort);
	for (i = 2; i < argc; ++i) {
		prfprintf(prfp, "%d) %s\n", i-1, argv[i]);
		prfflush(prfp);
	}
	prfclose(prfp);
	return 0;
}
#endif
